from __future__ import annotations

import ast
from pathlib import Path
from typing import Iterator


IMPORTS = "imports"
DEFINES = "defines"
ASSIGNS = "assigns"
SYM_TYPES = [IMPORTS, DEFINES, ASSIGNS]


def _imports(node: ast.AST):
    for alias in node.names:
        yield alias.name, IMPORTS


def _defines(node: ast.AST):
    yield node.name, DEFINES


def _assigns(node: ast.AST):
    for target in node.targets:
        if isinstance(target, ast.Name):
            yield target.id, ASSIGNS


SYM_MAP = {
    ast.Assign: _assigns,
    ast.FunctionDef: _defines,
    ast.ClassDef: _defines,
    ast.Import: _imports,
    ast.ImportFrom: _imports,
}


def iter_symbols(path: Path) -> Iterator[tuple[str, str]]:
    # sourcery skip: use-named-expression
    try:
        root = ast.parse(path.read_text("utf-8"))
    except SyntaxError:
        return

    for node in ast.iter_child_nodes(root):
        func = SYM_MAP.get(type(node))
        if func:
            yield from func(node)



from __future__ import annotations

import sqlite3
from pathlib import Path
from typing import Iterator

from symdexer.symbols import iter_symbols
from symdexer.modules import walk_modules, Module


class Cache:
    def __init__(self, path: Path):
        self.path = path

    def __enter__(self):
        self.db = sqlite3.connect(self.path).__enter__()
        return self

    def __exit__(self, *args, **kwargs):
        self.db.__exit__(*args, **kwargs)

    def init(self):
        self.db.executescript(
            """
            DROP TABLE IF EXISTS Module;

            DROP TABLE IF EXISTS Symbol;

            CREATE TABLE Module (
                name TEXT PRIMARY KEY,
                path TEXT NOT NULL,
                changed NUMBER NOT NULL
            );

            CREATE TABLE Symbol (
                name TEXT NOT NULL,
                type TEXT NOT NULL,
                module TEXT NOT NULL REFERENCES Module(name) ON UPDATE CASCADE,
                UNIQUE (name, type, module)
            );
            """
        )

    def update(self, packages: list[Path]) -> Iterator[Module]:
        for path in packages:
            for module in walk_modules(path):
                yield module
                self._index_module(module)

        self.db.commit()

    def _index_module(self, module: Module):
        if self.db.execute(
            """
            SELECT path, changed
            FROM module
            WHERE name = ? AND (path != ? OR changed != ?)
            """,
            module.info,
        ).fetchall():
            return

        self.db.execute(
            """
            INSERT OR REPLACE INTO Module (name, path, changed)
            VALUES (?, ? ,?)
            """,
            module.info,
        )

        for symbol, sym_type in iter_symbols(module.path):
            self.db.execute(
                """
                INSERT OR REPLACE INTO Symbol (name, type, module)
                VALUES (?, ? ,?)
                """,
                (symbol, sym_type, module.name),
            )

    def ungrouped(self, symbols: list[str], types: list[str], fuzzy: str) -> Iterator[tuple[str, str, str]]:
        if fuzzy:
            symbols = [f"%{s}%" for s in symbols]
            symbol_t = "Symbol.name LIKE ?"
        else:
            symbol_t = "Symbol.name = ?"

        symbols_t = " OR ".join(symbol_t for _ in symbols)
        types_t = " OR ".join("type = ?" for _ in types)

        cursor = self.db.execute(
            f"""
            SELECT Symbol.name, module, path
            FROM Symbol
            INNER JOIN Module ON module = Module.name
            WHERE ({symbols_t}) AND ({types_t})
            ORDER BY LENGTH(module) + LENGTH(Symbol.name) ASC
            """,
            (*symbols, *types),
        )

        return cursor.fetchall()

    def grouped(self, symbols: list[str], types: list[str], fuzzy: str) -> Iterator[tuple[str, str, str]]:
        if fuzzy:
            symbols = [f"%{s}%" for s in symbols]
            symbol_t = "Symbol.name LIKE ?"
        else:
            symbol_t = "Symbol.name = ?"

        symbols_t = " OR ".join(symbol_t for _ in symbols)
        types_t = " OR ".join("type = ?" for _ in types)

        cursor = self.db.execute(
            f"""
            SELECT GROUP_CONCAT(Symbol.name, ", ") as names, module, path
            FROM Symbol
            INNER JOIN Module ON module = Module.name
            WHERE ({symbols_t}) AND ({types_t})
            GROUP BY module
            ORDER BY LENGTH(module) + LENGTH(names) ASC
            """,
            (*symbols, *types),
        )

        return cursor.fetchall()



from __future__ import annotations

from typing import Iterator

from symdexer.cache import Cache
from symdexer.parser import Arguments, parse_args


def main():
    args = parse_args()

    if args.command == "index":
        cmd = index_command
    elif args.command == "import":
        cmd = import_command
    elif args.command == "locate":
        cmd = locate_command

    for line in cmd(args):
        print(line)


def index_command(args: Arguments) -> Iterator[str]:
    if args.reset and args.cache.is_file():
        args.cache.unlink()

    with Cache(args.cache) as cache:
        if not args.cache.exists():
            cache.init()

        for module in cache.update(args.packages):
            yield f"Indexing {module.name}"

    yield "Done indexing"


def import_command(args: Arguments) -> Iterator[str]:
    if not args.cache.is_file():
        raise FileNotFoundError(f"Cache `{args.cache}` not found.")

    with Cache(args.cache) as cache:
        search = cache.grouped if args.group else cache.ungrouped

        for name, module, _ in search(args.symbols, args.types, args.fuzzy):
            yield f"from {module} import {name}"


def locate_command(args: Arguments) -> Iterator[str]:
    if not args.cache.is_file():
        raise FileNotFoundError(f"Cache `{args.cache}` not found.")

    with Cache(args.cache) as cache:
        search = cache.grouped if args.group else cache.ungrouped

        for values in search(args.symbols, args.types, args.fuzzy):
            yield from values


if __name__ == "__main__":
    main()



from __future__ import annotations

import importlib
from pathlib import Path
from argparse import ArgumentParser, ArgumentTypeError, Namespace, _SubParsersAction

from symdexer.symbols import SYM_TYPES
from symdexer.__version__ import NAME, VERSION


class Arguments(Namespace):
    cache: Path
    command: str

    symbols: list[str]
    types: list[str]
    fuzzy: bool
    group: bool

    packages: list[Path]
    reset: bool


def parse_args() -> Arguments:
    parser = ArgumentParser(NAME)

    parser.add_argument("-v", "--version", action="version", version=VERSION)

    parser.add_argument(
        "-c",
        "--cache-file",
        dest="cache",
        metavar="file",
        default=Path("symdex.db"),
        type=Path,
        help="Name of the cache to store to and load from. Defaults to `symdex.db`",
    )

    sub_parsers = parser.add_subparsers(required=True, dest="command")

    _index_parser(sub_parsers)
    _import_parser(sub_parsers)
    _locate_parser(sub_parsers)

    return parser.parse_args()


def _index_parser(sub_parsers: _SubParsersAction[ArgumentParser]) -> None:
    parser = sub_parsers.add_parser(
        "index",
        help="Index symbols into cache",
    )

    parser.add_argument(
        "packages",
        metavar="PACKAGE",
        nargs="+",
        type=_package_type,
        help="The packages to add to cache. May be package names or paths",
    )
    parser.add_argument(
        "-r",
        "--reset",
        dest="reset",
        action="store_true",
        default=False,
        help="If the cache should be completely wiped",
    )


def _package_type(name: str) -> Path:
    path = Path(name)

    if path.exists():
        if path.is_dir() or (path.is_file() and path.suffix == ".py"):
            return path
        raise ArgumentTypeError(f"`{path}` is not a package")

    try:
        path = Path(importlib.import_module(name).__file__)
    except ImportError:
        raise ArgumentTypeError(f"`{name} is not a package`") from None

    if path.stem == "__init__":
        path = path.parent

    return path


def _import_parser(sub_parsers: _SubParsersAction[ArgumentParser]) -> None:
    parser = sub_parsers.add_parser(
        "import",
        help="Generates all possible imports for the matching symbols.",
    )

    _common_import_parser(parser)


def _locate_parser(sub_parsers: _SubParsersAction[ArgumentParser]) -> None:
    parser = sub_parsers.add_parser(
        "locate",
        help="Locates all possible imports for the matching symbols.",
    )

    _common_import_parser(parser)


def _common_import_parser(parser: ArgumentParser):
    parser.add_argument(
        "symbols",
        metavar="SYMBOL",
        nargs="+",
        help="The symbols to look for.",
    )
    parser.add_argument(
        "-t",
        "--types",
        dest="types",
        metavar="SYM_TYPE",
        choices=SYM_TYPES,
        default=SYM_TYPES,
        nargs="+",
        help="Symbol types to show.",
    )
    parser.add_argument(
        "-f",
        "--fuzzy",
        dest="fuzzy",
        action="store_true",
        default=False,
        help="Finds every symbol that are substrings of any of the provided symbols.",
    )
    parser.add_argument(
        "-g",
        "--group",
        dest="group",
        action="store_true",
        default=False,
        help="Groups symbols that come from the same package.",
    )



from __future__ import annotations

from pathlib import Path
from typing import Iterator
from functools import lru_cache
from dataclasses import dataclass


@dataclass(frozen=True)
class Module:
    path: Path
    name: str

    @property
    @lru_cache(maxsize=None)
    def path_str(self) -> str:
        return str(self.path.resolve())

    @property
    @lru_cache(maxsize=None)
    def mtime(self) -> int:
        return int(self.path.stat().st_mtime)

    @property
    @lru_cache(maxsize=True)
    def info(self) -> tuple[str, str, int]:
        return self.name, self.path_str, self.mtime


def walk_modules(path: Path, prefix: str = "") -> Iterator[Module]:
    if not prefix:
        prefix = path.stem

    if path.is_file():
        if path.suffix == ".py":
            yield Module(path, prefix)
        return

    for subpath in path.iterdir():
        yield from walk_modules(
            subpath,
            prefix if subpath.stem == "__init__" else f"{prefix}.{subpath.stem}",
        )
